@@ -15,14 +15,15 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */
+   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA. */
 
 
 #ifndef __CRC32_H__
 #define __CRC32_H__
 
 
-gboolean crc32_file_with_ID3_tag (gchar *filename, guint32 *crc32);
+gboolean crc32_file_with_ID3_tag (const gchar *filename, guint32 *crc32,
+                                  GError **err);
 
 
 #endif /* __CRC32_H__ */
