@@ -115,121 +115,84 @@ static void Ogg_Set_Tag (vorbis_comment *vc, const gchar *tag_name, gchar *value
  *************/
 
 /*
- * Read tag data into an Ogg Vorbis file.
- * Note:
- *  - if field is found but contains no info (strlen(str)==0), we don't read it
+ * convert_to_byte_array:
+ * @n: Integer to convert
+ * @array: Destination byte array
+ *
+ * Converts an integer to byte array.
  */
-gboolean
-ogg_tag_read_file_tag (gchar *filename, File_Tag *FileTag, GError **error)
+static void
+convert_to_byte_array (guint32 n, guchar *array)
 {
-    GFile *file;
-    GFileInputStream *istream;
-    vcedit_state   *state;
-    vorbis_comment *vc;
-    gchar          *string = NULL;
-    gchar          *string1 = NULL;
-    gchar          *string2 = NULL;
-    gchar          *filename_utf8 = filename_to_display(filename);
-    guint           field_num, i;
-    Picture        *prev_pic = NULL;
+    array [0] = (n >> 24) & 0xFF;
+    array [1] = (n >> 16) & 0xFF;
+    array [2] = (n >> 8) & 0xFF;
+    array [3] = n & 0xFF;
+}
 
-    g_return_val_if_fail (filename != NULL && FileTag != NULL, FALSE);
-    g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
+/*
+ * add_to_guchar_str:
+ * @ustr: Destination string
+ * @ustr_len: Pointer to length of destination string
+ * @str: String to append
+ * @str_len: Length of str
+ *
+ * Append a guchar string to given guchar string.
+ */
 
-    file = g_file_new_for_path (filename);
-    istream = g_file_read (file, NULL, error);
+static void
+add_to_guchar_str (guchar *ustr, gsize *ustr_len, guchar *str, gsize str_len)
+{
+    gsize i;
 
-    if (!istream)
+    for (i = *ustr_len; i < *ustr_len + str_len; i++)
     {
-        g_object_unref (file);
-        g_assert (error == NULL || *error != NULL);
-        return FALSE;
+        ustr [i] = str [i - *ustr_len];
     }
 
-    {
-    // Skip the id3v2 tag
-    guchar tmp_id3[4];
-    gulong id3v2size;
+    *ustr_len += str_len;
+}
 
-    // Check if there is an ID3v2 tag...
-    if (!g_seekable_seek (G_SEEKABLE (istream), 0L, G_SEEK_SET, NULL, error))
+/*
+ * read_guint_from_byte:
+ * @str: the byte string
+ * @start: position to start with
+ *
+ * Reads and returns an integer from given byte string starting from start.
+ * Returns: Integer which is read
+ */
+static guint32
+read_guint32_from_byte (guchar *str, gsize start)
+{
+    gsize i;
+    guint32 read = 0;
+
+    for (i = start; i < start + 4; i++)
     {
-        goto err;
+        read = (read << 8) + str[i];
     }
 
-    if (g_input_stream_read (G_INPUT_STREAM (istream), tmp_id3, 4, NULL,
-                             error) == 4)
-    {
-        // Calculate ID3v2 length
-        if (tmp_id3[0] == 'I' && tmp_id3[1] == 'D' && tmp_id3[2] == '3' && tmp_id3[3] < 0xFF)
-        {
-            // id3v2 tag skipeer $49 44 33 yy yy xx zz zz zz zz [zz size]
-            /* Size is 6-9 position */
-            if (!g_seekable_seek (G_SEEKABLE (istream), 2, G_SEEK_CUR,
-                                  NULL, error))
-            {
-                goto err;
-            }
+    return read;
+}
 
-            if (g_input_stream_read (G_INPUT_STREAM (istream), tmp_id3, 4,
-                                     NULL, error) == 4)
-            {
-                id3v2size = 10 + ( (long)(tmp_id3[3])        | ((long)(tmp_id3[2]) << 7)
-                                | ((long)(tmp_id3[1]) << 14) | ((long)(tmp_id3[0]) << 21) );
+/*
+ * et_add_file_tags_from_vorbis_comments:
+ * @vc: the byte string
+ * @FileTag: position to start with
+ * @filename_utf8: display filename
+ *
+ * Reads vorbis comments and copies them to file tag.
+ */
 
-                if (!g_seekable_seek (G_SEEKABLE (istream), id3v2size,
-                                      G_SEEK_SET, NULL, error))
-                {
-                    goto err;
-                }
-
-                Log_Print(LOG_ERROR,_("Warning: The Ogg Vorbis file '%s' contains an ID3v2 tag."),filename_utf8);
-            }
-            else if (!g_seekable_seek (G_SEEKABLE (istream), 0L, G_SEEK_SET,
-                                       NULL, error))
-            {
-                goto err;
-            }
-
-        }
-        else if (!g_seekable_seek (G_SEEKABLE (istream), 0L, G_SEEK_SET,
-                                   NULL, error))
-        {
-            goto err;
-        }
-
-    }
-    else if (!g_seekable_seek (G_SEEKABLE (istream), 0L, G_SEEK_SET,
-                               NULL, error))
-    {
-        goto err;
-    }
-
-    }
-
-    g_assert (error == NULL || *error == NULL);
-
-    g_object_unref (istream);
-
-    state = vcedit_new_state();    // Allocate memory for 'state'
-
-    if (!vcedit_open (state, file, error))
-    {
-        g_assert (error == NULL || *error != NULL);
-        g_object_unref (file);
-        vcedit_clear(state);
-        return FALSE;
-    }
-
-    g_assert (error == NULL || *error == NULL);
-
-    /* Get data from tag */
-    vc = vcedit_comments(state);
-    /*{
-        gint i; 
-        for (i=0;i<vc->comments;i++) 
-            g_print("%s -> Ogg vc:'%s'\n",g_path_get_basename(filename),vc->user_comments[i]);
-    }*/
+void
+et_add_file_tags_from_vorbis_comments (vorbis_comment *vc, File_Tag *FileTag,
+                                       const gchar *filename_utf8)
+{
+    gchar *string = NULL;
+    gchar *string1 = NULL;
+    gchar *string2 = NULL;
+    guint field_num, i;
+    Picture *prev_pic = NULL;
 
     /*********
      * Title *
@@ -548,6 +511,10 @@ ogg_tag_read_file_tag (gchar *filename, File_Tag *FileTag, GError **error)
     {
         Picture *pic;
             
+        /* Force marking the file as modified, so that the deprecated cover art
+         * field in converted in a METADATA_PICTURE_BLOCK field. */
+        FileTag->saved = FALSE;
+
         pic = Picture_Allocate();
         if (!prev_pic)
             FileTag->picture = pic;
@@ -577,6 +544,70 @@ ogg_tag_read_file_tag (gchar *filename, File_Tag *FileTag, GError **error)
 
     }
 
+    /* METADATA_BLOCK_PICTURE tag used for picture information */
+    field_num = 0;
+    while ((string = vorbis_comment_query (vc, "METADATA_BLOCK_PICTURE",
+                                           field_num++)) != NULL)
+    {
+        Picture *pic;
+        gsize bytes_pos, mimelen, desclen;
+        guchar *decoded_ustr;
+        gsize decoded_len;
+
+        pic = Picture_Allocate();
+
+        if (!prev_pic)
+        {
+            FileTag->picture = pic;
+        }
+        else
+        {
+            prev_pic->next = pic;
+        }
+
+        prev_pic = pic;
+
+        pic->data = NULL;
+
+        /* Decode picture data. */
+        decoded_ustr = g_base64_decode (string, &decoded_len);
+
+        /* Reading picture type. */
+        pic->type = read_guint32_from_byte (decoded_ustr, 0);
+        bytes_pos = 4;
+
+        /* Reading MIME data. */
+        mimelen = read_guint32_from_byte (decoded_ustr, bytes_pos);
+        bytes_pos = 8 + mimelen;
+
+        /* Reading description */
+        desclen = read_guint32_from_byte (decoded_ustr, bytes_pos);
+        bytes_pos += 4;
+
+        pic->description = g_malloc (desclen + 1);
+
+        for (i = 0; i < desclen; i++)
+        {
+            pic->description [i] = decoded_ustr [i + bytes_pos];
+        }
+
+        pic->description [desclen] = '\0';
+        bytes_pos += desclen + 16;
+
+        /* Reading picture size */
+        pic->size = read_guint32_from_byte (decoded_ustr, bytes_pos);
+        bytes_pos += 4;
+
+        /* Reading decoded picture */
+        pic->data = g_malloc (pic->size);
+
+        for (i = 0; i < pic->size; i++)
+        {
+            pic->data [i] = decoded_ustr [i + bytes_pos];
+        }
+
+        g_free (decoded_ustr);
+    }
 
     /***************************
      * Save unsupported fields *
@@ -604,13 +635,130 @@ ogg_tag_read_file_tag (gchar *filename, File_Tag *FileTag, GError **error)
           && strncasecmp(vc->user_comments[i],"COVERARTTYPE=",       13) != 0
           && strncasecmp(vc->user_comments[i],"COVERARTMIME=",       13) != 0
           && strncasecmp(vc->user_comments[i],"COVERARTDESCRIPTION=",20) != 0
+          && strncasecmp (vc->user_comments[i], "METADATA_BLOCK_PICTURE=", 23) != 0
            )
         {
             FileTag->other = g_list_append(FileTag->other,
                                            Try_To_Validate_Utf8_String(vc->user_comments[i]));
         }
     }
+}
 
+/*
+ * Read tag data into an Ogg Vorbis file.
+ * Note:
+ *  - if field is found but contains no info (strlen(str)==0), we don't read it
+ */
+gboolean
+ogg_tag_read_file_tag (gchar *filename, File_Tag *FileTag, GError **error)
+{
+    GFile *file;
+    GFileInputStream *istream;
+    vcedit_state   *state;
+    gchar *filename_utf8;
+
+    g_return_val_if_fail (filename != NULL && FileTag != NULL, FALSE);
+    g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
+
+    file = g_file_new_for_path (filename);
+    istream = g_file_read (file, NULL, error);
+
+    if (!istream)
+    {
+        g_object_unref (file);
+        g_assert (error == NULL || *error != NULL);
+        return FALSE;
+    }
+
+    filename_utf8 = filename_to_display (filename);
+
+    {
+    // Skip the id3v2 tag
+    guchar tmp_id3[4];
+    gulong id3v2size;
+
+    // Check if there is an ID3v2 tag...
+    if (!g_seekable_seek (G_SEEKABLE (istream), 0L, G_SEEK_SET, NULL, error))
+    {
+        goto err;
+    }
+
+    if (g_input_stream_read (G_INPUT_STREAM (istream), tmp_id3, 4, NULL,
+                             error) == 4)
+    {
+        // Calculate ID3v2 length
+        if (tmp_id3[0] == 'I' && tmp_id3[1] == 'D' && tmp_id3[2] == '3' && tmp_id3[3] < 0xFF)
+        {
+            // id3v2 tag skipeer $49 44 33 yy yy xx zz zz zz zz [zz size]
+            /* Size is 6-9 position */
+            if (!g_seekable_seek (G_SEEKABLE (istream), 2, G_SEEK_CUR,
+                                  NULL, error))
+            {
+                goto err;
+            }
+
+            if (g_input_stream_read (G_INPUT_STREAM (istream), tmp_id3, 4,
+                                     NULL, error) == 4)
+            {
+                id3v2size = 10 + ( (long)(tmp_id3[3])        | ((long)(tmp_id3[2]) << 7)
+                                | ((long)(tmp_id3[1]) << 14) | ((long)(tmp_id3[0]) << 21) );
+
+                if (!g_seekable_seek (G_SEEKABLE (istream), id3v2size,
+                                      G_SEEK_SET, NULL, error))
+                {
+                    goto err;
+                }
+
+                Log_Print(LOG_ERROR,_("Warning: The Ogg Vorbis file '%s' contains an ID3v2 tag."),filename_utf8);
+            }
+            else if (!g_seekable_seek (G_SEEKABLE (istream), 0L, G_SEEK_SET,
+                                       NULL, error))
+            {
+                goto err;
+            }
+
+        }
+        else if (!g_seekable_seek (G_SEEKABLE (istream), 0L, G_SEEK_SET,
+                                   NULL, error))
+        {
+            goto err;
+        }
+
+    }
+    else if (!g_seekable_seek (G_SEEKABLE (istream), 0L, G_SEEK_SET,
+                               NULL, error))
+    {
+        goto err;
+    }
+
+    }
+
+    g_assert (error == NULL || *error == NULL);
+
+    g_object_unref (istream);
+
+    state = vcedit_new_state();    // Allocate memory for 'state'
+
+    if (!vcedit_open (state, file, error))
+    {
+        g_assert (error == NULL || *error != NULL);
+        g_object_unref (file);
+        vcedit_clear(state);
+        g_free (filename_utf8);
+        return FALSE;
+    }
+
+    g_assert (error == NULL || *error == NULL);
+
+    /* Get data from tag */
+    /*{
+        gint i; 
+        for (i=0;i<vc->comments;i++) 
+            g_print("%s -> Ogg vc:'%s'\n",g_path_get_basename(filename),vc->user_comments[i]);
+    }*/
+
+    et_add_file_tags_from_vorbis_comments (vcedit_comments(state), FileTag,
+                                           filename_utf8);
     vcedit_clear(state);
     g_object_unref (file);
     g_free(filename_utf8);
@@ -679,7 +827,7 @@ ogg_tag_write_file_tag (ET_File *ETFile, GError **error)
     vcedit_state   *state;
     vorbis_comment *vc;
     gchar          *string;
-    GList          *list;
+    GList *l;
     Picture        *pic;
 
     g_return_val_if_fail (ETFile != NULL && ETFile->FileTag != NULL, FALSE);
@@ -861,51 +1009,147 @@ ogg_tag_write_file_tag (ET_File *ETFile, GError **error)
     /***********
      * Picture *
      ***********/
-    pic = FileTag->picture;
-    while (pic)
+    for (pic = FileTag->picture; pic != NULL; pic = pic->next)
     {
         if (pic->data)
         {
-            gchar *data_encoded = NULL;
-            Picture_Format format = Picture_Format_From_Data(pic);
+            const gchar *mime;
+            guchar array[4];
+            guchar *ustring = NULL;
+            gsize ustring_len = 0;
+            gchar *base64_string;
+            Picture_Format format = Picture_Format_From_Data (pic);
 
-            string = g_strdup_printf("COVERARTMIME=%s",Picture_Mime_Type_String(format));
-            vorbis_comment_add(vc,string);
-            g_free(string);
-        
-            if (pic->type)
+            /* According to the specification, only PNG and JPEG images should
+             * be added to Vorbis comments. */
+            if (format != PICTURE_FORMAT_PNG && format != PICTURE_FORMAT_JPEG)
             {
-                string = g_strdup_printf("COVERARTTYPE=%d",pic->type);
-                vorbis_comment_add(vc,string);
-                g_free(string);
+                GdkPixbufLoader *loader;
+                GError *error = NULL;
+
+                loader = gdk_pixbuf_loader_new ();
+
+                if (!gdk_pixbuf_loader_write (loader, pic->data, pic->size,
+                                              &error))
+                {
+                    Log_Print (LOG_ERROR, _("Error with 'loader_write': %s"),
+                               error->message);
+                    g_error_free (error);
+                    g_object_unref (loader);
+                    continue;
+                }
+                else
+                {
+                    GdkPixbuf *pixbuf;
+                    gchar *buffer;
+                    gsize buffer_size;
+
+                    if (!gdk_pixbuf_loader_close (loader, &error))
+                    {
+                        Log_Print (LOG_ERROR,
+                                   _("Error with 'loader_close': %s"),
+                                   error->message);
+                        g_error_free (error);
+                        g_object_unref (loader);
+                        continue;
+                    }
+
+                    pixbuf = gdk_pixbuf_loader_get_pixbuf (loader);
+
+                    if (!pixbuf)
+                    {
+                        g_object_unref (loader);
+                        continue;
+                    }
+
+                    g_object_ref (pixbuf);
+                    g_object_unref (loader);
+
+                    /* Always convert to PNG. */
+                    if (!gdk_pixbuf_save_to_buffer (pixbuf, &buffer,
+                                                    &buffer_size, "png",
+                                                    &error, NULL))
+                    {
+                        g_debug ("Error while converting image to PNG: %s",
+                                 error->message);
+                        g_error_free (error);
+                        g_object_unref (pixbuf);
+                        continue;
+                    }
+
+                    g_object_unref (pixbuf);
+
+                    g_free (pic->data);
+                    pic->data = (guchar *)buffer;
+                    pic->size = buffer_size;
+
+                    /* Set the picture format to reflect the new data. */
+                    format = Picture_Format_From_Data (pic);
+                }
             }
 
-            if (pic->description)
-            {
-                string = g_strdup_printf("COVERARTDESCRIPTION=%s",pic->description);
-                vorbis_comment_add(vc,string);
-                g_free(string);
-            }
+            mime = Picture_Mime_Type_String (format);
 
-            data_encoded = g_base64_encode (pic->data, pic->size);
-            string = g_strdup_printf("COVERART=%s",data_encoded);
-            vorbis_comment_add(vc,string);
-            g_free(data_encoded);
-            g_free(string);
+            /* Calculating full length of byte string and allocating. */
+            ustring = g_malloc (4 * 8 + strlen (mime)
+                                + strlen (pic->description) + pic->size);
+
+            /* Adding picture type. */
+            convert_to_byte_array (pic->type, array);
+            add_to_guchar_str (ustring, &ustring_len, array, 4);
+
+            /* Adding MIME string and its length. */
+            convert_to_byte_array (strlen (mime), array);
+            add_to_guchar_str (ustring, &ustring_len, array, 4);
+            add_to_guchar_str (ustring, &ustring_len, (guchar *)mime,
+                               strlen (mime));
+
+            /* Adding picture description. */
+            convert_to_byte_array (strlen (pic->description), array);
+            add_to_guchar_str (ustring, &ustring_len, array, 4);
+            add_to_guchar_str (ustring, &ustring_len,
+                               (guchar *)pic->description,
+                               strlen (pic->description));
+
+            /* Adding width, height, color depth, indexed colors. */
+            convert_to_byte_array (pic->width, array);
+            add_to_guchar_str (ustring, &ustring_len, array, 4);
+
+            convert_to_byte_array (pic->height, array);
+            add_to_guchar_str (ustring, &ustring_len, array, 4);
+
+            convert_to_byte_array (0, array);
+            add_to_guchar_str (ustring, &ustring_len, array, 4);
+
+            convert_to_byte_array (0, array);
+            add_to_guchar_str (ustring, &ustring_len, array, 4);
+
+            /* Adding picture data and its size. */
+            convert_to_byte_array (pic->size, array);
+            add_to_guchar_str (ustring, &ustring_len, array, 4);
+
+            add_to_guchar_str (ustring, &ustring_len, pic->data, pic->size);
+
+            base64_string = g_base64_encode (ustring, ustring_len);
+            string = g_strconcat ("METADATA_BLOCK_PICTURE=", base64_string,
+                                  NULL);
+            vorbis_comment_add (vc, string);
+
+            g_free (base64_string);
+            g_free (ustring);
+            g_free (string);
         }
-
-        pic = pic->next;
     }
 
     /**************************
      * Set unsupported fields *
      **************************/
-    list = FileTag->other;
-    while (list)
+    for (l = FileTag->other; l != NULL; l = g_list_next (l))
     {
-        if (list->data)
-            vorbis_comment_add(vc,(gchar *)list->data);
-        list = list->next;
+        if (l->data)
+        {
+            vorbis_comment_add (vc, (gchar *)l->data);
+        }
     }
 
     /* Write tag to 'file' in all cases */
