@@ -42,13 +42,26 @@
 
 #include "win32/win32dep.h"
 
+/* Data passed to et_picture_not_opened_callback */
+typedef struct
+{
+    GError *error;
+    gchar *filename;
+} ErrorDialogData;
+
+/* Data passed to et_picture_save_file_data_thread_func */
+typedef struct
+{
+    GFile *file;
+    Picture *pic;
+    GError *error;
+} PictureSaveData;
 
 /**************
  * Prototypes *
  **************/
 
 static void et_picture_load_file (GFile *file, gpointer user_data);
-
 static const gchar *Picture_Format_String (Picture_Format format);
 static const gchar *Picture_Type_String (EtPictureType type);
 static gchar *Picture_Info (Picture *pic);
@@ -69,12 +82,181 @@ static gboolean et_picture_save_file_data (const Picture *pic, GFile *file,
  * Functions *
  *************/
 
+/*
+ * et_picture_save_file_data_callback:
+ * @object: Source object of GAsyncResult
+ * @res: GAsyncResult for which callback is called
+ * @user_data: User data associated with GAsyncResult
+ *
+ * The callback function which is called when GSimpleAsyncResult of
+ * Picture_Save_Button_Clicked completes its operation.
+ */
+static void
+et_picture_save_file_data_callback (GObject *object, GAsyncResult *res,
+                                    gpointer user_data)
+{
+    PictureSaveData *data;
+
+    data = g_async_result_get_user_data (G_ASYNC_RESULT (res));
+
+    if (data->error)
+    {
+        Log_Print (LOG_ERROR, _("Image file not saved: %s"),
+                   data->error->message);
+        g_error_free (data->error);
+    }
+
+    gtk_widget_set_sensitive (MainWindow, TRUE);
+    g_slice_free (PictureSaveData, data);
+}
+
+/*
+ * et_picture_save_file_data_thread_func:
+ * @res: GSimpleAsyncResult associated with this function.
+ * @obj: Source object of GAsyncResult.
+ * @canc: Object to cancel the operation.
+ *
+ * Thread function of GSimpleAsyncResult of Picture_Save_Button_Clicked.
+ * This function will load pictures.
+ */
+static void
+et_picture_save_file_data_thread_func (GSimpleAsyncResult *res, GObject *obj,
+                                       GCancellable *cancellable)
+{
+    PictureSaveData *data;
+
+    data = g_async_result_get_user_data (G_ASYNC_RESULT (res));
+    et_picture_save_file_data (data->pic, data->file, &data->error);
+    g_object_unref (data->file);
+}
+
+/*
+ * et_picture_not_opened_callback:
+ * @object: Source object of GAsyncResult
+ * @res: GAsyncResult for which callback is called
+ * @user_data: User data associated with GAsyncResult
+ *
+ * The callback function which is called when et_picture_load_file
+ * reports an error.
+ */
+static void
+et_picture_not_opened_callback (GObject *object, GAsyncResult *res,
+                                gpointer user_data)
+{
+    GtkWidget *msgdialog;
+    ErrorDialogData *data = user_data;
+
+    /* Picture file not opened */
+    msgdialog = gtk_message_dialog_new (GTK_WINDOW (MainWindow),
+                                        GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
+                                        GTK_MESSAGE_ERROR,
+                                        GTK_BUTTONS_CLOSE,
+                                        _("Cannot open file: '%s'"),
+                                        data->filename);
+    gtk_message_dialog_format_secondary_text (GTK_MESSAGE_DIALOG(msgdialog),
+                                              "%s", data->error->message);
+    gtk_window_set_title (GTK_WINDOW (msgdialog), _("Image File Error"));
+    gtk_dialog_run (GTK_DIALOG (msgdialog));
+    gtk_widget_destroy (msgdialog);
+
+    Log_Print (LOG_ERROR, _("Image file not loaded: %s"),
+               data->error->message);
+
+    g_error_free (data->error);
+    g_free (data->filename);
+    g_slice_free (ErrorDialogData, data);
+}
+
+/*
+ * et_load_pictures_callback:
+ * @object: Source object of GAsyncResult
+ * @res: GAsyncResult for which callback is called
+ * @user_data: User data associated with GAsyncResult
+ *
+ * The callback function which is called when GSimpleAsyncResult of
+ * Picture_Add_Button_Clicked completes its operation.
+ */
+static void
+et_load_pictures_callback (GObject *object, GAsyncResult *res,
+                           gpointer user_data)
+{
+    GSList *list = g_async_result_get_user_data (res);
+
+    g_slist_free_full (list, g_object_unref);
+    gtk_widget_set_sensitive (MainWindow, TRUE);
+    g_object_unref (res);
+}
+
+/*
+ * et_load_pictures_thread_func:
+ * @res: GSimpleAsyncResult associated with this function.
+ * @obj: Source object of GAsyncResult.
+ * @canc: Object to cancel the operation.
+ *
+ * Thread function of GSimpleAsyncResult of Picture_Add_Button_Clicked.
+ * This function will load pictures.
+ */
+static void
+et_load_pictures_thread_func (GSimpleAsyncResult *res, GObject *obj,
+                              GCancellable *canc)
+{
+    GSList *list = g_async_result_get_user_data (G_ASYNC_RESULT (res));
+
+    g_slist_foreach (list, (GFunc) et_picture_load_file, NULL);
+}
+
+/*
+ * et_picture_load_file_callback:
+ * @source: Source object of GAsyncResult
+ * @res: GAsyncResult for which callback is called
+ * @user_data: User data associated with GAsyncResult
+ *
+ * The callback function which is called when simpleasync in
+ * Tag_Area_Picture_Drag_Data completes its operation.
+ */
+static void
+et_picture_load_file_callback (GObject *source, GAsyncResult *res,
+                               gpointer user_data)
+{
+    gtk_widget_set_sensitive (MainWindow, TRUE);
+    g_object_unref (res);
+}
+
+/*
+ * et_picture_load_file_thread_func:
+ * @res: GSimpleAsyncResult associated with this function.
+ * @obj: Source object of GAsyncResult.
+ * @cancellable: Object to cancel the operation.
+ *
+ * Thread function of simpleasync in Tag_Area_Picture_Drag_Data.
+ * This function will load pictures.
+ */
+static void
+et_picture_load_file_thread_func (GSimpleAsyncResult *res, GObject *obj,
+                                  GCancellable *cancellable)
+{
+    gchar **uri_list = g_async_result_get_user_data (G_ASYNC_RESULT (res));
+    gchar **uri = uri_list;
+
+    while (*uri && *(*uri))
+    {
+        GFile *file = g_file_new_for_uri (*uri);
+
+        et_picture_load_file (file, NULL);
+        g_object_unref (file);
+        uri++;
+    }
+
+    g_strfreev (uri_list);
+}
+
 void Tag_Area_Picture_Drag_Data (GtkWidget *widget, GdkDragContext *dc,
                                  gint x, gint y, GtkSelectionData *selection_data,
                                  guint info, guint t, gpointer data)
 {
     GtkTreeSelection *selection;
-    gchar **uri_list, **uri;
+    gchar **uri_list;
+    GSimpleAsyncResult *simpleasync;
 
     gtk_drag_finish(dc, TRUE, FALSE, t);
 
@@ -86,17 +268,15 @@ void Tag_Area_Picture_Drag_Data (GtkWidget *widget, GdkDragContext *dc,
     selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(PictureEntryView));
     gtk_tree_selection_unselect_all(selection);
 
-    uri = uri_list = g_strsplit((gchar *)gtk_selection_data_get_data(selection_data), "\r\n", 0);
-    while (*uri && strlen(*uri))
-    {
-        GFile *file = g_file_new_for_uri (*uri);
-
-        et_picture_load_file (file, NULL);
-
-        g_object_unref (file);
-        uri++;
-    }
-    g_strfreev(uri_list);
+    uri_list = g_strsplit((gchar *)gtk_selection_data_get_data(selection_data), "\r\n", 0);
+    simpleasync = g_simple_async_result_new (NULL,
+                                             et_picture_load_file_callback,
+                                             uri_list,
+                                             Tag_Area_Picture_Drag_Data);
+    gtk_widget_set_sensitive (MainWindow, FALSE);
+    g_simple_async_result_run_in_thread (simpleasync,
+                                         et_picture_load_file_thread_func,
+                                         G_PRIORITY_DEFAULT, NULL);
 }
 
 void
@@ -251,24 +431,14 @@ et_picture_load_file (GFile *file, gpointer user_data)
 
     if (!pic)
     {
-        GtkWidget *msgdialog;
-
-        /* Picture file not opened */
-        msgdialog = gtk_message_dialog_new (GTK_WINDOW (MainWindow),
-                                            GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
-                                            GTK_MESSAGE_ERROR,
-                                            GTK_BUTTONS_CLOSE,
-                                            _("Cannot open file: '%s'"),
-                                            filename_utf8);
-        gtk_message_dialog_format_secondary_text (GTK_MESSAGE_DIALOG(msgdialog),
-                                                  "%s", error->message);
-        gtk_window_set_title (GTK_WINDOW (msgdialog), _("Image File Error"));
-        gtk_dialog_run (GTK_DIALOG (msgdialog));
-        gtk_widget_destroy (msgdialog);
-
-        Log_Print (LOG_ERROR, _("Image file not loaded: %s"),
-                   error->message);
-        g_error_free (error);
+        GSimpleAsyncResult *simple;
+        ErrorDialogData *data = g_slice_new (ErrorDialogData);
+        data->error = error;
+        data->filename = g_strdup (filename_utf8);
+        simple = g_simple_async_result_new (NULL, et_picture_not_opened_callback,
+                                            data, NULL);
+        g_simple_async_result_complete_in_idle (simple);
+        g_object_unref (info);
         return;
     }
     else
@@ -289,6 +459,7 @@ et_picture_load_file (GFile *file, gpointer user_data)
             // Other tag types
             case ID3_TAG:
             case OGG_TAG:
+            case OPUS_TAG:
             case APE_TAG:
             case FLAC_TAG:
             case WAVPACK_TAG:
@@ -301,10 +472,9 @@ et_picture_load_file (GFile *file, gpointer user_data)
                 g_assert_not_reached ();
         }
 
-        PictureEntry_Update(pic, TRUE);
+        PictureEntry_Update (pic, TRUE);
 
-        // FIXME: Call Picture_Free(pic) here? It seems PictureEntry_Update makes copies of pic.
-        //Picture_Free(pic);
+        Picture_Free (pic);
     }
 
     g_object_unref (info);
@@ -394,14 +564,18 @@ void Picture_Add_Button_Clicked (GObject *object)
     {
         GtkTreeSelection *selection;
         GSList *list;
+        GSimpleAsyncResult *res;
 
         selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(PictureEntryView));
         gtk_tree_selection_unselect_all(selection);
 
         list = gtk_file_chooser_get_files (GTK_FILE_CHOOSER (FileSelectionWindow));
-        g_slist_foreach (list, (GFunc) et_picture_load_file, NULL);
-        g_slist_free_full (list, g_object_unref);
-
+        res = g_simple_async_result_new (NULL, et_load_pictures_callback,
+                                         list, Picture_Add_Button_Clicked);
+        gtk_widget_set_sensitive (MainWindow, FALSE);
+        g_simple_async_result_run_in_thread (res,
+                                             et_load_pictures_thread_func,
+                                             G_PRIORITY_DEFAULT, NULL);
         // Save the directory selected for initialize next time
         g_free(init_dir);
         init_dir = gtk_file_chooser_get_current_folder(GTK_FILE_CHOOSER(FileSelectionWindow));
@@ -484,6 +658,8 @@ void Picture_Properties_Button_Clicked (GObject *object)
                                                          NULL);
         g_free(title);
 
+        gtk_window_set_default_size (GTK_WINDOW (PictureTypesWindow), 400,
+                                     400);
         gtk_dialog_set_default_response(GTK_DIALOG(PictureTypesWindow), GTK_RESPONSE_OK);
 
         ScrollWindowPictureTypes = gtk_scrolled_window_new(NULL, NULL);
@@ -503,7 +679,6 @@ void Picture_Properties_Button_Clicked (GObject *object)
                                             "text", PICTURE_TYPE_COLUMN_TEXT,
                                             NULL);
         gtk_tree_view_append_column(GTK_TREE_VIEW(type), column);
-        gtk_widget_set_size_request(type, 256, 256);
         gtk_box_pack_start(GTK_BOX(gtk_dialog_get_content_area(GTK_DIALOG(PictureTypesWindow))),ScrollWindowPictureTypes,TRUE,TRUE,0);
 
         // Behaviour following the tag type...
@@ -729,6 +904,9 @@ void Picture_Save_Button_Clicked (GObject *object)
                 case PICTURE_FORMAT_PNG :
                     image_name = g_strdup("image_name.png");
                     break;
+                case PICTURE_FORMAT_GIF:
+                    image_name = g_strdup ("image_name.gif");
+                    break;
                 case PICTURE_FORMAT_UNKNOWN :
                     image_name = g_strdup("image_name.ext");
                     break;
@@ -746,22 +924,24 @@ void Picture_Save_Button_Clicked (GObject *object)
         if (response == GTK_RESPONSE_OK)
         {
             GFile *file;
-            GError *error = NULL;
+            GSimpleAsyncResult *res;
+            PictureSaveData *data = g_slice_new (PictureSaveData);
 
             // Save the directory selected for initialize next time
             g_free(init_dir);
             init_dir = gtk_file_chooser_get_current_folder(GTK_FILE_CHOOSER(FileSelectionWindow));
 
             file = gtk_file_chooser_get_file (GTK_FILE_CHOOSER (FileSelectionWindow));
-
-            if (!et_picture_save_file_data (pic, file, &error))
-            {
-                 Log_Print (LOG_ERROR, _("Image file not saved: %s"),
-                            error->message);
-                 g_error_free (error);
-            }
-
-            g_object_unref (file);
+            gtk_widget_set_sensitive (MainWindow, FALSE);
+            data->file = file;
+            data->pic = pic;
+            data->error = NULL;
+            res = g_simple_async_result_new (NULL,
+                                             et_picture_save_file_data_callback,
+                                             data, NULL);
+            g_simple_async_result_run_in_thread (res,
+                                                 et_picture_save_file_data_thread_func,
+                                                 G_PRIORITY_DEFAULT, NULL);
         }
         gtk_widget_destroy(FileSelectionWindow);
     }
@@ -791,7 +971,7 @@ Picture_Format Picture_Format_From_Data (Picture *pic)
     &&  pic->data[7] == 0x0a)
         return PICTURE_FORMAT_PNG;
     
-    /*// GIF : "GIF87a"
+    /* GIF: "GIF87a" */
     if (pic->data && pic->size > 6
     &&  pic->data[0] == 0x47
     &&  pic->data[1] == 0x49
@@ -799,7 +979,19 @@ Picture_Format Picture_Format_From_Data (Picture *pic)
     &&  pic->data[3] == 0x38
     &&  pic->data[4] == 0x37
     &&  pic->data[5] == 0x61)
-        return PICTURE_FORMAT_GIF;*/
+        return PICTURE_FORMAT_GIF;
+
+    /* GIF: "GIF89a" */
+    if (pic->data && pic->size > 6
+        &&  pic->data[0] == 0x47
+        &&  pic->data[1] == 0x49
+        &&  pic->data[2] == 0x46
+        &&  pic->data[3] == 0x38
+        &&  pic->data[4] == 0x39
+        &&  pic->data[5] == 0x61)
+    {
+        return PICTURE_FORMAT_GIF;
+    }
     
     return PICTURE_FORMAT_UNKNOWN;
 }
@@ -812,8 +1004,11 @@ const gchar *Picture_Mime_Type_String (Picture_Format format)
             return "image/jpeg";
         case PICTURE_FORMAT_PNG:
             return "image/png";
+        case PICTURE_FORMAT_GIF:
+            return "image/gif";
         default:
-            return NULL;
+            g_debug ("%s", "Unrecognised image MIME type");
+            return "application/octet-stream";
     }
 }
 
@@ -827,6 +1022,8 @@ Picture_Format_String (Picture_Format format)
             return _("JPEG image");
         case PICTURE_FORMAT_PNG:
             return _("PNG image");
+        case PICTURE_FORMAT_GIF:
+            return _("GIF image");
         default:
             return _("Unknown image");
     }
