@@ -66,6 +66,13 @@ static void vcedit_clear_internals(vcedit_state *state)
         vorbis_info_clear(state->vi);
         free(state->vi);
     }
+#ifdef ENABLE_OPUS
+    if (state->oi)
+    {
+        g_free (state->oi);
+    }
+#endif /* ENABLE_OPUS */
+
     g_object_unref (state->in);
     memset(state, 0, sizeof(*state));
 }
@@ -104,6 +111,12 @@ static int _commentheader_out(vcedit_state *state, ogg_packet *op)
         oggpack_write(&opb,0x03,8);
         _v_writestring(&opb,"vorbis", 6);
     }
+#ifdef ENABLE_OPUS
+    else if (state->oggtype == VCEDIT_IS_OPUS)
+    {
+        _v_writestring (&opb, "OpusTags", 8);
+    }
+#endif
 
     /* vendor */
     oggpack_write(&opb,strlen(vendor),32);
@@ -351,16 +364,29 @@ vcedit_open(vcedit_state *state, GFile *file, GError **error)
     {
         state->oggtype = VCEDIT_IS_OGGVORBIS;
     }
-#ifdef ENABLE_SPEEX
     else
     {
+#ifdef ENABLE_SPEEX
         // Done after "Ogg test" to avoid to display an error message in function
         // speex_packet_to_header when the file is not Speex.
         state->si = malloc(sizeof(SpeexHeader));
         if((state->si = speex_packet_to_header((char*)(&header_main)->packet,(&header_main)->bytes)))
             state->oggtype = VCEDIT_IS_SPEEX;
-    }
 #endif
+#ifdef ENABLE_OPUS
+        if (state->oggtype == VCEDIT_IS_UNKNOWN)
+        {
+            state->oi = g_malloc (sizeof (OpusHead));
+
+            if (opus_head_parse (state->oi,
+                                 (unsigned char*)(&header_main)->packet,
+                                 (&header_main)->bytes) == 0)
+            {
+                state->oggtype = VCEDIT_IS_OPUS;
+            }
+        }
+#endif
+    }
 
     if (state->oggtype == VCEDIT_IS_UNKNOWN)
     {
@@ -380,6 +406,11 @@ vcedit_open(vcedit_state *state, GFile *file, GError **error)
             header = &header_comments;
             headerpackets = 2 + state->si->extra_headers;
             break;
+#endif
+#ifdef ENABLE_OPUS
+        case VCEDIT_IS_OPUS:
+            header = &header_comments;
+            headerpackets = 2;
 #endif
     }
     i = 1;
@@ -434,6 +465,35 @@ vcedit_open(vcedit_state *state, GFile *file, GError **error)
                                     goto err;
                                     break;
                             }
+                            break;
+#ifdef ENABLE_OPUS
+                        case VCEDIT_IS_OPUS:
+                            switch (opus_tags_parse ((OpusTags *)state->vc,
+                                                     header->packet,
+                                                     header->bytes))
+                            {
+                                case OP_ENOTFORMAT:
+                                    g_set_error (error, ET_OGG_ERROR,
+                                                 ET_OGG_ERROR_HEADER,
+                                                 "Ogg Opus tags do not start with \"OpusTags\"");
+                                    goto err;
+                                    break;
+
+                                case OP_EFAULT:
+                                    g_set_error (error, ET_OGG_ERROR,
+                                                 ET_OGG_ERROR_HEADER,
+                                                 "Not enough memory to store Ogg Opus tags");
+                                    goto err;
+                                    break;
+
+                                case OP_EBADHEADER:
+                                    g_set_error (error, ET_OGG_ERROR,
+                                                 ET_OGG_ERROR_HEADER,
+                                                 "Ogg Opus tags do not follow the specification");
+                                    goto err;
+                                    break;
+                            }
+#endif
                     }
                     i++;
                 }
@@ -487,19 +547,27 @@ vcedit_write(vcedit_state *state, GFile *file, GError **error)
     char *buffer;
     int bytes;
     int needflush=0, needout=0;
-    GFileOutputStream *ostream;
+    GOutputStream *ostream;
+    gchar *buf;
+    gsize size;
+    GFileInfo *fileinfo;
 
     g_return_val_if_fail (error == NULL || *error == NULL, FALSE);
 
-    ostream = g_file_replace (file, NULL, FALSE, G_FILE_CREATE_NONE,
-                              NULL, error);
-
-    if (!ostream)
+    fileinfo = g_file_query_info (file, G_FILE_ATTRIBUTE_STANDARD_SIZE,
+                                  G_FILE_QUERY_INFO_NONE, NULL, error);
+    if (!fileinfo)
     {
         g_assert (error == NULL || *error != NULL);
         return FALSE;
     }
 
+    buf = g_malloc (g_file_info_get_size (fileinfo));
+    ostream = g_memory_output_stream_new (buf,
+                                          g_file_info_get_size (fileinfo),
+                                          g_realloc, g_free);
+    g_object_unref (fileinfo);
+
     state->eosin = 0;
     state->extrapage = 0;
 
@@ -522,16 +590,16 @@ vcedit_write(vcedit_state *state, GFile *file, GError **error)
     ogg_stream_packetin(&streamout, &header_main);
     ogg_stream_packetin(&streamout, &header_comments);
 
-    if (state->oggtype != VCEDIT_IS_SPEEX)
+    if (state->oggtype == VCEDIT_IS_OGGVORBIS)
         ogg_stream_packetin(&streamout, &header_codebooks);
 
     while((result = ogg_stream_flush(&streamout, &ogout)))
     {
         gsize bytes_written;
 
-        if (!g_output_stream_write_all (G_OUTPUT_STREAM (ostream),
-                                        ogout.header, ogout.header_len,
-                                        &bytes_written, NULL, error))
+        if (!g_output_stream_write_all (ostream, ogout.header,
+                                        ogout.header_len, &bytes_written, NULL,
+                                        error))
         {
             g_debug ("Only %" G_GSIZE_FORMAT " bytes out of %ld bytes of data "
                      "were written", bytes_written, ogout.header_len);
@@ -539,9 +607,8 @@ vcedit_write(vcedit_state *state, GFile *file, GError **error)
             goto cleanup;
         }
 
-        if (!g_output_stream_write_all (G_OUTPUT_STREAM (ostream), ogout.body,
-                                        ogout.body_len, &bytes_written, NULL,
-                                        error))
+        if (!g_output_stream_write_all (ostream, ogout.body, ogout.body_len,
+                                        &bytes_written, NULL, error))
         {
             g_debug ("Only %" G_GSIZE_FORMAT " bytes out of %ld bytes of data "
                      "were written", bytes_written, ogout.body_len);
@@ -554,26 +621,29 @@ vcedit_write(vcedit_state *state, GFile *file, GError **error)
     {
         if(needflush)
         {
-            gsize bytes_written;
-
-            if (!g_output_stream_write_all (G_OUTPUT_STREAM (ostream),
-                                            ogout.header, ogout.header_len,
-                                            &bytes_written, NULL, error))
+            if (ogg_stream_flush (&streamout, &ogout))
             {
-                g_debug ("Only %" G_GSIZE_FORMAT " bytes out of %ld bytes of "
-                         "data were written", bytes_written, ogout.header_len);
-                g_assert (error == NULL || *error != NULL);
-                goto cleanup;
-            }
+                gsize bytes_written;
 
-            if (!g_output_stream_write_all (G_OUTPUT_STREAM (ostream),
-                                            ogout.body, ogout.body_len,
-                                            &bytes_written, NULL, error))
-            {
-                g_debug ("Only %" G_GSIZE_FORMAT " bytes out of %ld bytes of "
-                         "data were written", bytes_written, ogout.body_len);
-                g_assert (error == NULL || *error != NULL);
-                goto cleanup;
+                if (!g_output_stream_write_all (ostream, ogout.header,
+                                                ogout.header_len, &bytes_written,
+                                                NULL, error))
+                {
+                    g_debug ("Only %" G_GSIZE_FORMAT " bytes out of %ld bytes of "
+                             "data were written", bytes_written, ogout.header_len);
+                    g_assert (error == NULL || *error != NULL);
+                    goto cleanup;
+                }
+
+                if (!g_output_stream_write_all (ostream, ogout.body,
+                                                ogout.body_len, &bytes_written,
+                                                NULL, error))
+                {
+                    g_debug ("Only %" G_GSIZE_FORMAT " bytes out of %ld bytes of "
+                             "data were written", bytes_written, ogout.body_len);
+                    g_assert (error == NULL || *error != NULL);
+                    goto cleanup;
+                }
             }
         }
         else if(needout)
@@ -582,8 +652,8 @@ vcedit_write(vcedit_state *state, GFile *file, GError **error)
             {
                 gsize bytes_written;
 
-                if (!g_output_stream_write_all (G_OUTPUT_STREAM (ostream),
-                                                ogout.header, ogout.header_len,
+                if (!g_output_stream_write_all (ostream, ogout.header,
+                                                ogout.header_len,
                                                 &bytes_written, NULL, error))
                 {
                     g_debug ("Only %" G_GSIZE_FORMAT " bytes out of %ld bytes "
@@ -593,9 +663,9 @@ vcedit_write(vcedit_state *state, GFile *file, GError **error)
                     goto cleanup;
                 }
 
-                if (!g_output_stream_write_all (G_OUTPUT_STREAM (ostream),
-                                                ogout.body, ogout.body_len,
-                                                &bytes_written, NULL, error))
+                if (!g_output_stream_write_all (ostream, ogout.body,
+                                                ogout.body_len, &bytes_written,
+                                                NULL, error))
                 {
                     g_debug ("Only %" G_GSIZE_FORMAT " bytes out of %ld bytes "
                              "of data were written", bytes_written,
@@ -608,12 +678,22 @@ vcedit_write(vcedit_state *state, GFile *file, GError **error)
 
         needflush=needout=0;
 
-        if (state->oggtype == VCEDIT_IS_OGGVORBIS)
+        if (state->oggtype == VCEDIT_IS_OGGVORBIS ||
+            state->oggtype == VCEDIT_IS_OPUS)
         {
-            int size;
-            size = _blocksize(state, &op);
-            granpos += size;
-
+            if (state->oggtype == VCEDIT_IS_OGGVORBIS)
+            {
+                int size;
+                size = _blocksize(state, &op);
+                granpos += size;
+            }
+#ifdef ENABLE_OPUS
+            else
+            {
+                granpos += opus_packet_get_samples_per_frame (op.packet,
+                                                              48000);
+            }
+#endif
             if(op.granulepos == -1)
             {
                 op.granulepos = granpos;
@@ -664,9 +744,9 @@ vcedit_write(vcedit_state *state, GFile *file, GError **error)
     {
         gsize bytes_written;
 
-        if (!g_output_stream_write_all (G_OUTPUT_STREAM (ostream),
-                                        ogout.header, ogout.header_len,
-                                        &bytes_written, NULL, error))
+        if (!g_output_stream_write_all (ostream, ogout.header,
+                                        ogout.header_len, &bytes_written, NULL,
+                                        error))
         {
             g_debug ("Only %" G_GSIZE_FORMAT " bytes out of %ld bytes of data "
                      "were written", bytes_written, ogout.header_len);
@@ -674,9 +754,8 @@ vcedit_write(vcedit_state *state, GFile *file, GError **error)
             goto cleanup;
         }
 
-        if (!g_output_stream_write_all (G_OUTPUT_STREAM (ostream), ogout.body,
-                                        ogout.body_len, &bytes_written, NULL,
-                                        error))
+        if (!g_output_stream_write_all (ostream, ogout.body, ogout.body_len,
+                                        &bytes_written, NULL, error))
         {
             g_debug ("Only %" G_GSIZE_FORMAT " bytes out of %ld bytes of data "
                      "were written", bytes_written, ogout.body_len);
@@ -689,9 +768,9 @@ vcedit_write(vcedit_state *state, GFile *file, GError **error)
     {
         gsize bytes_written;
 
-        if (!g_output_stream_write_all (G_OUTPUT_STREAM (ostream),
-                                        ogout.header, ogout.header_len,
-                                        &bytes_written, NULL, error))
+        if (!g_output_stream_write_all (ostream, ogout.header,
+                                        ogout.header_len, &bytes_written, NULL,
+                                        error))
         {
             g_debug ("Only %" G_GSIZE_FORMAT " bytes out of %ld bytes of data "
                      "were written", bytes_written, ogout.header_len);
@@ -699,9 +778,8 @@ vcedit_write(vcedit_state *state, GFile *file, GError **error)
             goto cleanup;
         }
 
-        if (!g_output_stream_write_all (G_OUTPUT_STREAM (ostream), ogout.body,
-                                        ogout.body_len, &bytes_written, NULL,
-                                        error))
+        if (!g_output_stream_write_all (ostream, ogout.body, ogout.body_len,
+                                        &bytes_written, NULL, error))
         {
             g_debug ("Only %" G_GSIZE_FORMAT " bytes out of %ld bytes of data "
                      "were written", bytes_written, ogout.body_len);
@@ -732,17 +810,17 @@ vcedit_write(vcedit_state *state, GFile *file, GError **error)
 
                 /* Don't bother going through the rest, we can just
                  * write the page out now */
-                if (!g_output_stream_write_all (G_OUTPUT_STREAM (ostream),
-                                                ogout.header, ogout.header_len,
+                if (!g_output_stream_write_all (ostream, ogout.header,
+                                                ogout.header_len,
                                                 &bytes_written, NULL, error))
                 {
                     g_assert (error == NULL || *error != NULL);
                     goto cleanup;
                 }
 
-                if (!g_output_stream_write_all (G_OUTPUT_STREAM (ostream),
-                                                ogout.body, ogout.body_len,
-                                                &bytes_written, NULL, error))
+                if (!g_output_stream_write_all (ostream, ogout.body,
+                                                ogout.body_len, &bytes_written,
+                                                NULL, error))
                 {
                     g_assert (error == NULL || *error != NULL);
                     goto cleanup;
@@ -786,12 +864,39 @@ cleanup:
         }
     }
 
-    g_object_unref (ostream);
-
     if (error == NULL || *error != NULL)
+    {
+        g_object_unref (ostream);
         return FALSE;
+    }
 
     g_assert (error == NULL || *error == NULL);
+
+    if (!g_output_stream_close (ostream, NULL, error))
+    {
+        g_object_unref (ostream);
+        g_assert (error == NULL || *error != NULL);
+        return FALSE;
+    }
+
+    g_assert (error == NULL || *error == NULL);
+
+    buf = g_memory_output_stream_steal_data (G_MEMORY_OUTPUT_STREAM (ostream));
+    size = g_memory_output_stream_get_data_size (G_MEMORY_OUTPUT_STREAM (ostream));
+
+    /* Write the in-memory data back out to the original file. */
+    if (!g_file_replace_contents (file, buf, size, NULL, FALSE,
+                                  G_FILE_CREATE_NONE, NULL, NULL, error))
+    {
+       g_object_unref (ostream);
+       g_free (buf);
+       g_assert (error == NULL || *error != NULL);
+       return FALSE;
+    }
+
+    g_free (buf);
+    g_object_unref (ostream);
+
     return TRUE;
 }
 
